classdef moleculeData < handle
  
  properties
    moleculeNumber
    traceNum
    fileName
    numFrames
    
   analysisStepHistory %cell array to keep track of steps that are taken and order of steps
    originalData %struct with fields: donor, acceptor, fret
    kmeansData 
    separatedStates
    nameOfExperiment = [];
    donorZero = 0;
    acceptorZero = 0;
    leakage = 0;
    % later on, read about sparce matrices because this may be the best way
    % to store the data: https://www.mathworks.com/help/matlab/sparse-matrices.html
    
  end
  properties (SetObservable = true)
    currentData
    currentAnalysisStep  (1,:) char {mustBeMember(currentAnalysisStep, ...
      {'originalData', 'leakageApplied', 'zeroAdjusted', 'dataFilterApplied','redFramesRemoved','postClustering', 'firstAndLastRemoved', ...
      'singleFramesAnalyzed'})} = 'originalData';
  end
  events
    updateCurrentAnalysisStep
    oops
    removeFirstAndLastClusters
    applyDataFilter
    applyLeakageSettings
    adjustZero
  end
  
  methods %constructor function
    function mol = moleculeData(m, data, handleToApp, nameOfExperiment)
      if nargin > 0
        if nargin == 4
          mol.nameOfExperiment = nameOfExperiment;
        end
      mol.moleculeNumber = m;
      mol.traceNum = data.traceNum;
      mol.fileName = data.traceName;
      originalData.donor = data.donor; originalData.acceptor = data.acceptor; 
      originalData.fret = data.fret;
      mol.originalData = originalData;
      mol.currentData = originalData;
      mol.numFrames = length(mol.originalData.donor);
      mol.analysisStepHistory.stepsTaken = {'originalData'};
      mol.analysisStepHistory.stepInfo = {[]};
      mol.analysisStepHistory.changesToData = {[]};
      % event listeners
      %addlistener(handleToApp, 'removeBleachingAndBlinking', @(src,evnt)mol.removeBleachedSections(src,evnt));
      %addlistener(mol, 'removeBleaching', @(src,evnt)mol.removeBleachedSections(evnt));
      addlistener(handleToApp, 'updateRedFrames', @(src,evnt) mol.removeOrReplaceRedFrameDataPoints(evnt));
      addlistener(mol, 'oops', @(src,evnt) mol.undoLastStep);
      addlistener(mol, 'updateCurrentAnalysisStep', @(s,evnt) mol.updateAnalysisStepHistory(evnt));
      addlistener(mol, 'removeFirstAndLastClusters', @(s,e) mol.removeFirstAndLastSections);
      addlistener(mol, 'applyLeakageSettings', @(s,e) mol.applyLeakage(e));
      addlistener(mol, 'adjustZero', @(s,e) mol.adjustZeroForData(e));
      addlistener(mol, 'applyDataFilter', @(s,e) mol.filterData(e));
      end
      end
  end
  
  methods 

    function updateAnalysisStepHistory(mol,evnt)
      arguments
        mol
        evnt = [];
      end
      if isempty(evnt)
        mol.analysisStepHistory.stepsTaken{end+1} = mol.currentAnalysisStep;
        mol.analysisStepHistory.stepInfo{end+1} = struct();
        mol.analysisStepHistory.changesToData{end+1} = struct();
        return
      end
      stepInfo = evnt.data;
      if isfield(stepInfo, 'currentAnalysisStep')
        mol.currentAnalysisStep = stepInfo.currentAnalysisStep;
        stepInfo = rmfield(stepInfo, 'currentAnalysisStep');
      end
      stepsSoFar = mol.analysisStepHistory.stepsTaken;
      col = length(stepsSoFar) + 1;
      switch mol.currentAnalysisStep
        case {'redFramesRemoved', 'zeroAdjusted', 'leakageApplied'}
          stepsSoFar = mol.analysisStepHistory.stepsTaken;
          col = getAnalysisStepColumnToUpdate(stepsSoFar, mol.currentAnalysisStep);
        otherwise
%           mol.analysisStepHistory.stepInfo{end+1} = stepInfo;
%           mol.analysisStepHistory.stepsTaken{end+1} = mol.currentAnalysisStep;
%           mol.analysisStepHistory.changesToData{end+1} = evnt.changesToData;
      end
      mol.analysisStepHistory.stepInfo{col} = stepInfo;
      mol.analysisStepHistory.stepsTaken{col} = mol.currentAnalysisStep;
      mol.analysisStepHistory.changesToData{col} = evnt.changesToData;
      function col = getAnalysisStepColumnToUpdate(stepsSoFar, currentStep)
          [~, col] = find(contains(stepsSoFar, currentStep),1, 'last');
          if isempty(col)
            col = length(stepsSoFar) + 1;
          end
      end
    end
    function applyLeakage(mol, evnt)
      %leakage = 0.00;			%for cy3 leakage into cy5 factor 0.09, set to zero for calcein/cy5
      %leakage = 0.052;   %41708 subtract bg in matlab %121307 BR
      %leakage = 0.073;   % CC,GT DNA cy3-cy5
      %leakage = 0.064;   % CC,GT DNA Tamra-cy5
      %leakage = 0.12;   % Alex555
            stepsSoFar = mol.analysisStepHistory.stepsTaken;
      [~, col] = find(contains(stepsSoFar, 'leakageApplied'),1, 'last');
      if ~isempty(col)
        %need to undo previous leakage first
        oldLeakage = mol.analysisStepHistory.stepInfo{col}.leakage;
        mol.currentData.acceptor = mol.currentData.acceptor + oldLeakage*mol.currentData.donor;
        mol.currentData.fret = mol.currentData.acceptor./(mol.currentData.acceptor + mol.currentData.donor);
      end
      mol.leakage = evnt.data.leakage;
      if mol.leakage ~= 0
      mol.currentData.acceptor = mol.currentData.acceptor - mol.leakage*mol.currentData.donor;
      mol.currentData.fret = mol.currentData.acceptor./(mol.currentData.acceptor + mol.currentData.donor);
      end
      notify(mol, 'updateCurrentAnalysisStep', evnt.data);
    end
    function removeOrReplaceRedFrameDataPoints(mol,evnt)
      oldData = mol.currentData;
      greenStartFrame = evnt.data.greenStartFrame;
      greenEndFrame = evnt.data.greenEndFrame;
      dataNames = {'fret'; 'donor'; 'acceptor'};
      switch evnt.data.source
        case 'startFrames'
          for i = 1:size(dataNames,1)
            %restore the current data to what it was
            mol.currentData.(dataNames{i})(1:evnt.data.oldValue) = mol.originalData.(dataNames{i})(1:evnt.data.oldValue);
            if evnt.data.removeTF == true %then set new frames to nan
              mol.currentData.(dataNames{i})(1:evnt.data.newValue) = nan;
            end
          end
        case 'endFrames'
          for i = 1:size(dataNames,1)
            %restore the current data to what it was
            if evnt.data.oldValue ~= inf
            mol.currentData.(dataNames{i})(evnt.data.oldValue:end) = mol.originalData.(dataNames{i})(evnt.data.oldValue:end);
            end
            if evnt.data.removeTF == true && evnt.data.newValue ~= inf %then set new frames to nan
              mol.currentData.(dataNames{i})(evnt.data.newValue:end) = nan;
            end
          end
        case 'checkBox'
          switch evnt.data.removeTF
            case true
              for i = 1:size(dataNames,1)
                mol.currentData.(dataNames{i})(1:greenStartFrame) = nan;
                if greenEndFrame ~= inf
                mol.currentData.(dataNames{i})(greenEndFrame:end) = nan;
                end
              end
            case false
              for i = 1:size(dataNames,1)
              mol.currentData.(dataNames{i})(1:greenStartFrame) = mol.originalData.(dataNames{i})(1:greenStartFrame);
              if greenEndFrame ~= inf
                mol.currentData.(dataNames{i})(greenEndFrame:end) = mol.originalData.(dataNames{i})(greenEndFrame:end);
              end
              end
          end
      end
      newEventData.valueThatChanged = evnt.data.source;
      newEventData.greenStartFrame = greenStartFrame;
      newEventData.greenEndFrame = greenEndFrame;
      newEventData.removeTF = evnt.data.removeTF;

      newEventData.currentAnalysisStep = 'redFramesRemoved';
      evntData = eventSpecificData(newEventData, oldData, mol.currentData);
      notify(mol, 'updateCurrentAnalysisStep', evntData);
    end
    
    function adjustZeroForData(mol, evnt)
      % if zero was already set, undo last settings first
      stepsSoFar = mol.analysisStepHistory.stepsTaken;
      [~, col] = find(contains(stepsSoFar, 'zeroAdjusted'),1, 'last');
      if ~isempty(col)
        oldAcceptorZero = mol.analysisStepHistory.stepInfo{col}.acceptorZeroAdjustment;
        oldDonorZero = mol.analysisStepHistory.stepInfo{col}.donorZeroAdjustment;
        mol.currentData.donor = mol.currentData.donor + oldDonorZero;
        mol.currentData.acceptor = mol.currentData.acceptor + oldAcceptorZero;
        mol.currentData.fret = mol.currentData.acceptor./(mol.currentData.acceptor + mol.currentData.donor);
      end
      mol.acceptorZero =  evnt.data.acceptorZeroAdjustment;
      mol.donorZero = evnt.data.donorZeroAdjustment;
      oldData = mol.currentData;
      mol.currentData.donor = oldData.donor - evnt.data.donorZeroAdjustment;
      mol.currentData.acceptor = oldData.acceptor - evnt.data.acceptorZeroAdjustment;
      mol.currentData.fret = mol.currentData.acceptor./(mol.currentData.acceptor + mol.currentData.donor);
      evntData = eventSpecificData(evnt.data, oldData, mol.currentData);
      notify(mol, 'updateCurrentAnalysisStep', evntData);
    end
    function filterData(mol, evnt)
      filterInputs = evnt.data;
      filter = dataFilter(mol.currentData, filterInputs);
      oldData = mol.currentData;
      mol.currentData = filter.filteredData;
      mol.currentAnalysisStep = 'dataFilterApplied';
      evntData = eventSpecificData(filterInputs, oldData, mol.currentData);
      notify(mol, 'updateCurrentAnalysisStep', evntData);
    end
    function removeBleachedSections(mol,evnt)
      % applies data filter settings to molecule and updates molecule step
      % history
      oldData = mol.currentData;
      fieldsToRemove = cell.empty; fieldsToAdd = cell.empty; 
      valuesOfFieldsToAdd = cell.empty;
      mol.currentAnalysisStep = 'postBleachRemoval';
      newEventData = evnt.data;
      %this is where the filtering happens
      switch evnt.data.method
        case {'removedByHand', 'autoFilterWithAbsoluteMedianDeviation'}
          bleachRemovalData = bleachRemovalResults(mol.currentData, evnt.data, 'removedByHand');
          if isfield(mol.currentData, 'clusteredSignal')
            bleachRemovalData.results.clusteredSignal = bleachRemovalData.results.acceptor;
            mol.currentData = bleachRemovalData.results;
            mol.currentAnalysisStep = 'postBleachRemoval';
          else
            mol.currentData = bleachRemovalData.results;
          end
          dataForEvent.method = evnt.data.method;
          evntData = eventSpecificData(dataForEvent, oldData, mol.currentData);
          notify(mol, 'updateCurrentAnalysisStep', evntData);
          return;
        case 'threshold if Red Frames Are Bleached'
          switch newEventData.dataToUse
            case 'Sum Acceptor Donor'
              redFrameData = mol.originalData.donor + mol.originalData.acceptor;
            case {'Acceptor', 'Donor', 'Fret'}
              dataName = lower(newEventData.dataToUse);
              redFrameData = mol.originalData.(dataName);
            case {'Donor AND Acceptor', 'Donor OR Acceptor'}
              redFrameData.donor = mol.originalData.donor;
              redFrameData.acceptor = mol.originalData.acceptor;
          end
          bleachRemovalData = bleachRemovalResults(mol.currentData, evnt.data, 'threshold if Red Frames Are Bleached', 'redFrameData', redFrameData);
        otherwise
          bleachRemovalData = bleachRemovalResults(mol.currentData, evnt.data);
      end
      % from here to the end is just modifying the analysis step history so
      % that it prints out correctly when it's exported.
      if bleachRemovalData.applyMaxThreshTF == 0
        fieldsToRemove{end+1} ='applyMaxThreshTF';
        fieldsToRemove{end+1} ='maxThresh';
        if ismember(bleachRemovalData.dataToUse, {'Donor And Acceptor Separately'})
          fieldsToRemove{end+1} ='threshold';
          fieldsToAdd{end+1} = 'acceptor threshold:';
          valuesOfFieldsToAdd{end+1} = bleachRemovalData.threshold;
          fieldsToAdd{end+1} = 'donor threshold:';
          valuesOfFieldsToAdd{end+1} = bleachRemovalData.maxThresh;
        end
%         for i = 1:3
%         fieldsToRemove{end+1} = moreFieldsToRemove{i}; %#ok
%         end
      end
      if bleachRemovalData.windowSize == 0
        fieldsToRemove{end+1} = 'windowSize';
        fieldsToRemove{end+1} = 'smoothFunction';
      end
      switch evnt.data.actionToTake
        case 'Remove'
          fieldsToRemove{end+1} = 'maxLengthOfBlink';
        case 'Replace WITH'
          fieldsToRemove{end+1} = 'maxLengthOfBlink';
          evnt.data.valueToReplaceWith = evnt.data.maxLengthOfBlink; %change the name
        case 'Remove IF'
         if ~isempty(bleachRemovalData.conditions)&& ismember(bleachRemovalData.conditions, {'Sections Before And After Are Bleached'})
           fieldsToRemove{end+1} = 'maxLengthOfBlink';
         end
                  if ~isempty(bleachRemovalData.conditions)&& ismember(evnt.data.method, {'threshold if Red Frames Are Bleached'})
           fieldsToRemove{end+1} = 'maxLengthOfBlink';
           fieldsToAdd{end+1} = 'lengthOfLaserSwitch';
          valuesOfFieldsToAdd{end+1} = evnt.data.maxLengthOfBlink;
         end
      end
      if isempty(bleachRemovalData.conditions)
        fieldsToRemove{end+1} = 'conditions';
      end
      if ~isempty(fieldsToRemove)
        for i = 1:size(fieldsToRemove,2)
          if isfield(newEventData, (fieldsToRemove{i}))
            newEventData = rmfield(newEventData, (fieldsToRemove{i}));
          end
        end
      end
      if ~isempty(fieldsToAdd)
        for i = 1:size(fieldsToAdd,2)
          newEventData.(fieldsToAdd{i}) = valuesOfFieldsToAdd{i};
        end
      end
      if isfield(mol.currentData, 'clusteredSignal')
        cs = mol.currentData.clusteredSignal;
        cs(bleachRemovalResults.bleachIndex) = nan;
        bleachRemovalData.results.clusteredSignal = cs;
        mol.currentData = bleachRemovalData.results;
        mol.currentAnalysisStep = 'postClustering';
      else
        mol.currentData = bleachRemovalData.results;
        mol.currentAnalysisStep = 'postBleachRemoval';
      end
      
    end
    function assignKmeansResults(mol, clusterIndex, kmeansInfo) %clusterIndex, k, centers, methodForMeasuringDistance, whichDataWasClustered)
      clusteredSignal = nan(mol.numFrames,1);
      clusterIndex = logical(clusterIndex);
      %kmeansInfo % struct with: k, centers, methodForMeasuringDistance, ...
      %whichDataWasClustered,whichMols, molNums, if data was smoothed then
      %also windowSize
      centers = kmeansInfo.centers;
      %signalInTermsOfClusterNumber = clusteredSignal;
      for j = 1:length(centers)
        clusteredSignal(clusterIndex(:,j)) = centers(j);
        %signalInTermsOfClusterNumber(clusterIndex(:,j)) = j;
      end
      oldData = mol.currentData;
      mol.currentData.clusteredSignal = clusteredSignal;
      %mol.kmeansData.signalInTermsOfClusterNumber = signalInTermsOfClusterNumber;
      mol.kmeansData.centers = centers;
      mol.kmeansData.k = kmeansInfo.k;
      mol.currentAnalysisStep = 'postClustering';
      if mol.kmeansData.k == 1
        newUpdateInfo = [];
        newUpdateInfo.stats = kmeansInfo.stats;
        newUpdateInfo.centers = kmeansInfo.centers;
        evntData = eventSpecificData(newUpdateInfo, oldData, mol.currentData);
      else
        evntData = eventSpecificData(kmeansInfo, oldData, mol.currentData);
      end
      
      notify(mol, 'updateCurrentAnalysisStep', evntData);
    end


    function removeFirstAndLastSections(mol)
      if all(isnan(mol.currentData.clusteredSignal))
        mol.currentAnalysisStep = 'firstAndLastRemoved';
        dataForEvent = 'firstAndLastSectionsRemoved';
        evntData = eventSpecificData(dataForEvent, mol.currentData, mol.currentData);
        notify(mol, 'updateCurrentAnalysisStep', evntData);
        return
      end

      clusterNumbers = mol.getSignalInTermsOfClusterNumber;
      % find the 'nan' sections - we want to remove the first and last
      % cluster of each stretch of data in between 'nan' values
      
      
      %%%%%%%%% use this instead
      cs = mol.currentData.clusteredSignal;
      idx = logical(isnan(cs));
      b = [idx(2:end); nan];
      [rowEnd, ~] = find(b ~= idx);
      c = [nan; idx(1:end-1)];
      [rowStart, ~] = find(c~= idx);
      % row start and row end are the start and end rows of the signal
      % sectioned into either nan or not nan. 
      if isnan(cs(rowStart(1)))
        startingNumber = 2;
      else
        startingNumber = 1;
      end
      sectionsToRemove = zeros(size(mol.currentData.clusteredSignal), 'logical');
      for i = startingNumber:2:length(rowStart)
        signalAfterSectionStart = cs(rowStart(i));
        idx = logical(cs == signalAfterSectionStart);
        b = [idx(2:end); nan];
        [sectionEnd, ~] = find(b ~= idx);
        c = [nan; idx(1:end-1)];
        [sectionStart, ~] = find(c~= idx);
        [r,~] = find(sectionStart == rowStart(i));
        sectionsToRemove(sectionStart(r):sectionEnd(r)) = true;
        %sectionToRemove{end+1} = sectionStart(r):sectionEnd(r);
        signalAfterSectionEnd = cs(rowEnd(i));
        idx = logical(cs == signalAfterSectionEnd);
        b = [idx(2:end); nan];
        [sectionEnd, ~] = find(b ~= idx);
        c = [nan; idx(1:end-1)];
        [sectionStart, ~] = find(c~= idx);
        [r,~] = find(sectionEnd == rowEnd(i));
        sectionsToRemove(sectionStart(r):sectionEnd(r)) = true;
        %sectionToRemove{end+1} = sectionStart(r):sectionEnd(r);
      end
     

      

      %%%%%%%%%%%%%%%%%%%%%%%%%%
%       
%       sectionsToRemove = zeros(size(clusterNumbers), 'logical');
%       sectionsToRemove(isnan(clusterNumbers)) = true;
%       [rows,~] = find(~isnan(clusterNumbers));
%       i = 1; sectionStart = rows(1); idx = 1; sections = cell.empty;
%       %first separate out sections between 'nan' values
%       while i < length(rows)
%         if rows(i+1) == rows(i) + 1
%           i = i+1;
%           if i == length(rows)
%             sections{idx} = sectionStart:rows(i);
%           end
%         else
%           sections{idx} = sectionStart:rows(i);
%           idx = idx + 1;
%           i = i+1;
%           sectionStart = rows(i);
%         end
%       end
%       %loop through sections and find the first and last cluster of each
%       numSections = length(sections);
%       for i = 1:numSections
%         section = sections{i};
%         clusterNumbersInSection = clusterNumbers(section);
%         if all(clusterNumbersInSection == clusterNumbersInSection(1))
%           sectionsToRemove(section) = true;
%         else
%           clusterNumberOfFirstClusterInSection = clusterNumbers(section(1));
%           [rowsOfSection, ~] = find(clusterNumbers(section)==clusterNumberOfFirstClusterInSection);
%           if length(rowsOfSection) == 1
%             sectionsToRemove(section(rowsOfSection)) = true;
%           else
%             j = 1;
%             while j < length(rowsOfSection)
%               if rowsOfSection(j+1) == rowsOfSection(j) + 1
%                 j = j+1;
%               else
%                 sectionsToRemove(section(1:rowsOfSection(j))) = true;
%                 j = length(rowsOfSection);
%               end
%             end
%           end
%           clusterNumberOfLastClusterInSection = clusterNumbers(section(end));
%           [rowsOfSection, ~] = find(clusterNumbers(section)==clusterNumberOfLastClusterInSection);
%           j = length(rowsOfSection);
%           while j > 1
%             if rowsOfSection(j-1) == rowsOfSection(j) - 1
%               j = j-1;
%             else
%               idx = rowsOfSection(j):rowsOfSection(end);
%               sectionsToRemove(section(idx)) = true;
%               j = 0;
%             end
%             if j == 1
%               sectionsToRemove(section(rowsOfSection)) = true;
%             end
%           end
%         end
%       end
      %remove first and last sections
      oldData = mol.currentData;
      dataNames = {'acceptor', 'donor', 'fret', 'clusteredSignal'};
      for i = 1:4
        data = oldData.(dataNames{i});
        data(sectionsToRemove) = nan;
        newCurrentData.(dataNames{i}) = data;
      end
      mol.currentData = newCurrentData;
      mol.currentAnalysisStep = 'firstAndLastRemoved';
      dataForEvent = 'firstAndLastSectionsRemoved';
      evntData = eventSpecificData(dataForEvent, oldData, mol.currentData);
      notify(mol, 'updateCurrentAnalysisStep', evntData);
    end
    

    function analyzeSingleFrameEvents(mol, clusterAvgs, methodToUse)
      sections = mol.separateOutSections(mol.currentData.clusteredSignal);
      [singleFrameEvents, consecutiveSingleFrameEvents] = mol.locateSingleFrameEvents(sections);
      %analyze consecutive single frame events first
      if ~isempty(consecutiveSingleFrameEvents)
      newClusterNumberAssignments = mol.analyzeConsecutiveSingleFrameEvents(consecutiveSingleFrameEvents);
      % n x 2 cell with column1 = rows to modify, and column2 = new value
      % to replace rows with.
      else
        newClusterNumberAssignments = [];
      end
      %analyze non-consecutive events next
      if ~isempty(singleFrameEvents)
        analysisMethod = methodToUse.method;
        switch analysisMethod
          case 'Donor & Acceptor'
            dataForEvent.analysisMethod = analysisMethod;
            newClusterNumberAssignments2 = mol.analyzeNonConsecutiveSingleFrameEvents(singleFrameEvents, clusterAvgs, sections);
          case 'avgd Fret'
            
            newClusterNumberAssignments2 = mol.analyzeNonConsecutiveSingleFrameEvents2(singleFrameEvents, methodToUse.numFrames);
        end
      else
        newClusterNumberAssignments2 = [];
      end
      %return updated signal
      oldData = mol.currentData;
      newAssignments = [newClusterNumberAssignments; newClusterNumberAssignments2];
      if ~isempty(newAssignments)
      dataNames = {'acceptor', 'donor', 'fret', 'clusteredSignal'};
      newCurrentData = mol.currentData; 
        for j = 1:size(newAssignments,1)
          frames = newAssignments{j,1};
          newValue = newAssignments{j,2};
          if isnan(newValue)
            for i = 1:4
              newCurrentData.(dataNames{i})(frames) = nan;
            end
          else
            %only need to update clustered signal
            newCurrentData.clusteredSignal(frames) = newValue;
          end
        end
      mol.currentData = newCurrentData;
      end
      mol.currentAnalysisStep = 'singleFramesAnalyzed';
%       dataForEvent.numberOfSingleFramesAnalyzed = length(singleFrameEvents) + length(consecutiveSingleFrameEvents);
%       dataForEvent.numberThatWereConsecutive = length(consecutiveSingleFrameEvents);
      dataForEvent = struct();
      switch methodToUse.method
        case 'Donor & Acceptor'
          dataForEvent.analysisMethod = methodToUse.method;
        case 'avgd Fret'
          dataForEvent.analysisMethod = methodToUse.method;
          dataForEvent.windowSize = methodToUse.numFrames;
      end
      
      evntData = eventSpecificData(dataForEvent, oldData, mol.currentData);
      notify(mol, 'updateCurrentAnalysisStep', evntData);
    end
    function newClusterNumberAssignments = analyzeConsecutiveSingleFrameEvents(mol, consecutiveFrames)
      %separate frames into sections of consecutive single frame events and
      %analyze each section as a group
      i = 1; sections = cell.empty; startFrame = consecutiveFrames(1);
      idx = 1;
      while i < length(consecutiveFrames)
        if consecutiveFrames(i+1) ~= consecutiveFrames(i) + 1
          sections{idx} = startFrame:consecutiveFrames(i);
          startFrame = consecutiveFrames(i+1);
          idx = idx + 1;
        end
        i = i+1;
      end
      sections{idx} = startFrame:consecutiveFrames(end);
      newClusterNumberAssignments = cell(length(sections),2);
      newClusterNumberAssignments(:,1) = sections;
      fretCenters = mol.kmeansData.centers;
      for i = 1:length(sections)
        % calculate average fret of the consecutive frames
        fretAvgValue = mean(mol.currentData.fret(sections{i}));
        %compare avg fret value to each cluster center and assign to
        %closest center
        distances = abs(fretCenters - fretAvgValue);
        [~,newClusterNumber] = min(distances);
        newClusterNumberAssignments{i,2} = fretCenters(newClusterNumber);
        %mol.currentData.clusteredSignal(sections{i}) = newClusterNumber;
        %mol.kmeansData.signalInTermsOfClusterNumber(sections{i}) = newClusterNumber;
      end
    end
    
    function newClusterNumberAssignments = analyzeNonConsecutiveSingleFrameEvents2(mol, frames, numToAvg)
      % for each frame, average the fret value at that frame with the fret
      % values in the frames immediately before and after this frame. Treat
      % the averaged value as the value for the single frame and re-assign it to
      % a cluster based on this value.
      newClusterNumberAssignments = cell(length(frames),2);
      fretCenters = mol.kmeansData.centers;
      numToAvgOnEachSide = floor((numToAvg - 1)/2);
      for i = 1:length(frames)
        newClusterNumberAssignments{i,1} = frames(i);
        frameIndex = (frames(i) - numToAvgOnEachSide):1:(frames(i) + numToAvgOnEachSide);
        %frameIndex = [frames(i)-1, frames(i),frames(i)+1];
        TF = frameIndex > 0 & frameIndex <= mol.numFrames;
        frameIndex = frameIndex(TF);
        TF = ~isnan(mol.currentData.fret(frameIndex));
        frameIndex = frameIndex(TF);
        numValuesToAvg = length(frameIndex);
        if numValuesToAvg > 1
          fretAvgValue = mean(mol.currentData.fret(frameIndex));
          distances = abs(fretCenters - fretAvgValue);
          [~,newClusterNumber] = min(distances);
          newClusterNumberAssignments{i,2} = fretCenters(newClusterNumber);
        else %remove it if it's in between nan values or is first or last in the trace
          newClusterNumberAssignments{i,2} = nan;
        end
      end
    end
    
    function signalInTermsOfClusterNumbers = getSignalInTermsOfClusterNumber(mol)
      centers = mol.kmeansData.centers;
      clusteredSignal = mol.currentData.clusteredSignal;
      signalInTermsOfClusterNumbers = nan(size(clusteredSignal));
      for i = 1:size(centers)
        indexForThisCluster = clusteredSignal == centers(i);
        signalInTermsOfClusterNumbers(indexForThisCluster) = i;
      end
    end
    
    
    function newClusterNumberAssignments = analyzeNonConsecutiveSingleFrameEvents(mol, singleEvents, clusterAvgs, sections)
      numSingleFrameEvents = length(singleEvents);
      newClusterNumberAssignments = cell(numSingleFrameEvents,2);
      newClusterNumberAssignments(:,1) = num2cell(singleEvents);
%       %make note of the first and last clusters in the trace
      [rowNumOfLastCluster,~] = find(~isnan(mol.currentData.clusteredSignal),1,'last');
      [rowNumOfFirstCluster,~] = find(~isnan(mol.currentData.clusteredSignal),1);
     
      clusterNumbers = mol.getSignalInTermsOfClusterNumber;
      [nanRows, ~] = find(isnan(clusterNumbers));
      
      dataNames = {'acceptor', 'donor', 'fret'};
      columnCorrespondingToDataNamesForClusterAvgVariable = [2, 3, 1];

      for i = 1:numSingleFrameEvents
        currentCluster.frames = singleEvents(i);
        % if it's next to a nan value, remove it
        rowsBeforeAndAfterSingleFrame = [currentCluster.frames-1, currentCluster.frames+1];
        if any(ismember(rowsBeforeAndAfterSingleFrame, nanRows))
          newClusterNumberAssignments{i,2} = nan;
          continue
        end
        
        % if it's the first or last cluster in the trace, remove it
        if ismember(currentCluster.frames, [rowNumOfFirstCluster, rowNumOfLastCluster])
          newClusterNumberAssignments{i,2} = nan;
          continue
        end
        % otherwise, compare current cluster to the left and right adjacent
        % clusters. 
        currentCluster.clusterNum = clusterNumbers(currentCluster.frames);
        leftCluster.clusterNum = clusterNumbers(currentCluster.frames-1);
        rightCluster.clusterNum = clusterNumbers(currentCluster.frames+1);
        % take note of frames for left and right clusters
        for j = 1:length(sections)
          if ismember(leftCluster.clusterNum, sections{j})
            leftCluster.frames = sections{j};
          end
          if ismember(rightCluster.clusterNum, sections{j})
            rightCluster.frames = sections{j};
          end
        end
        if leftCluster.clusterNum == rightCluster.clusterNum
          % then it looks like _/\_  combine the left and right and
          % see if it is closer to the adjacent cluster or the
          % average of the cluster it is already in. If it is closer to
          % the adjacent cluster, reassign it, otherwise, leave it
          % alone
          combinedAdjacentFrames = [leftCluster.frames, rightCluster.frames];
          adjacentCluster.frames = combinedAdjacentFrames;
          adjacentCluster.clusterNum = leftCluster.clusterNum;
          % combine adjacent cluster info
          for j = 1:2
            dataForCalculatingAvgs = mol.currentData.(dataNames{j})(combinedAdjacentFrames);
            adjacentCluster.(dataNames{j}).localAvg = mean(dataForCalculatingAvgs);
            adjacentCluster.(dataNames{j}).localStd = std(dataForCalculatingAvgs);

            adjacentCluster.(dataNames{j}).globalAvg = clusterAvgs.median(adjacentCluster.clusterNum, columnCorrespondingToDataNamesForClusterAvgVariable(j)); %switching from mean to median
            adjacentCluster.(dataNames{j}).globalStd = clusterAvgs.standardDeviation(adjacentCluster.clusterNum, columnCorrespondingToDataNamesForClusterAvgVariable(j));
            currentCluster.(dataNames{j}).localAvg = mol.currentData.(dataNames{j})(currentCluster.frames);
            currentCluster.(dataNames{j}).globalAvg = clusterAvgs.median(currentCluster.clusterNum, columnCorrespondingToDataNamesForClusterAvgVariable(j));
            currentCluster.(dataNames{j}).globalStd = clusterAvgs.standardDeviation(currentCluster.clusterNum, columnCorrespondingToDataNamesForClusterAvgVariable(j));
            
            % remove section below once section above is tested
%             fieldnameForGlobalAvgs = "cluster" + num2str(adjacentCluster.clusterNum);
%             adjacentCluster.(dataNames{j}).globalAvg = clusterAvgs.(fieldnameForGlobalAvgs).(dataNames{j}).avg;
%             adjacentCluster.(dataNames{j}).globalStd = clusterAvgs.(fieldnameForGlobalAvgs).(dataNames{j}).sd;
%             currentCluster.(dataNames{j}).localAvg = mol.currentData.(dataNames{j})(currentCluster.frames);
%             fieldnameForGlobalAvgs = "cluster" + num2str(currentCluster.clusterNum);
%             currentCluster.(dataNames{j}).globalAvg = clusterAvgs.(fieldnameForGlobalAvgs).(dataNames{j}).avg;
%             currentCluster.(dataNames{j}).globalStd = clusterAvgs.(fieldnameForGlobalAvgs).(dataNames{j}).sd;
          end
          % decide what cluster is closest
          keepSingleFrameEventInCurrentCluster = true;
          for j = 1:2
            if keepSingleFrameEventInCurrentCluster == true
              intensityValueAtSingleFrame = currentCluster.(dataNames{j}).localAvg;
              howCloseIsItToTheCurrentClustersGlobalAvg = abs(intensityValueAtSingleFrame - currentCluster.(dataNames{j}).globalAvg);
              howCloseIsItToTheAdjacentClustersGlobalAvg = abs(intensityValueAtSingleFrame - adjacentCluster.(dataNames{j}).globalAvg);
              howCloseIsItToTheAdjacentClustersLocalAvg = abs(intensityValueAtSingleFrame - adjacentCluster.(dataNames{j}).localAvg);
              closestDistanceToAdjacentCluster = min(howCloseIsItToTheAdjacentClustersGlobalAvg, howCloseIsItToTheAdjacentClustersLocalAvg);
              isTheCurrentClusterCloser = howCloseIsItToTheCurrentClustersGlobalAvg < closestDistanceToAdjacentCluster;
              if isTheCurrentClusterCloser == true
                howMuchCloser = abs(howCloseIsItToTheCurrentClustersGlobalAvg - closestDistanceToAdjacentCluster);
                %why am I comparing the amount closer to the standard
                %deviation? Maybe this should be changed..
                isTheAmountCloserMoreThanTheStdOfTheAdjacentCluster = howMuchCloser > adjacentCluster.(dataNames{j}).globalStd;
                if isTheAmountCloserMoreThanTheStdOfTheAdjacentCluster == true
                  keepSingleFrameEventInCurrentCluster = true;
                else
                  keepSingleFrameEventInCurrentCluster = false;
                end
              else
                keepSingleFrameEventInCurrentCluster = false;
              end
            end
          end
          
          switch keepSingleFrameEventInCurrentCluster
            case true
              % if acceptor and donor are both closer to the current
              % cluster, then don't do anything. Otherwise, reassign to
              % the adjacent cluster
              newClusterNumberAssignments{i,2} =  mol.kmeansData.centers(currentCluster.clusterNum);
            case false
              newClusterNumberAssignments{i,2} = mol.kmeansData.centers(adjacentCluster.clusterNum);
          end
          continue
        end %end of if leftCluster.clusterNum == rightCluster.clusterNum
        
        if leftCluster.clusterNum ~= rightCluster.clusterNum
          leftTest = 0; rightTest = 0;
          vars = [leftTest, rightTest];
          clusterNums = [leftCluster.clusterNum, rightCluster.clusterNum];
          dataNames = {'acceptor', 'donor'};
          for j = 1:2
%             fieldnameForGlobalAvgs = "cluster" + num2str(clusterNums(j));
            for q = 1:2
              avgGlobally = clusterAvgs.median(clusterNums(j),columnCorrespondingToDataNamesForClusterAvgVariable(j));
              valueAtSingleFrameEvent = mol.currentData.(dataNames{q})(currentCluster.frames);
              vars(j) = vars(j) + abs(avgGlobally - valueAtSingleFrameEvent);
            end
          end
          if rightTest < leftTest
            newClusterNumberAssignments{i,2} = mol.kmeansData.centers(rightCluster.clusterNum);
          else
            newClusterNumberAssignments{i,2} = mol.kmeansData.centers(leftCluster.clusterNum);
          end
        end %end of if leftCluster.clusterNum ~= right
      end %end of loop on all single frames
    end
  

    function undoLastStep(mol)
      stepToUndo = mol.analysisStepHistory.stepsTaken{end};
      dataUpdates = mol.analysisStepHistory.changesToData{end};
      if length(mol.analysisStepHistory.stepsTaken) > 1
      switch stepToUndo
        case 'postClustering'
          stepBeforeThisOne = mol.analysisStepHistory.stepsTaken{end-1};
          switch stepBeforeThisOne
            case 'postClustering'
              mol.kmeansData = mol.analysisStepHistory.stepInfo{end-1};
            otherwise
              mol.currentData = rmfield(mol.currentData, 'clusteredSignal');
          mol.kmeansData = [];
          end
        case 'leakageApplied'
          oldLeakage = mol.analysisStepHistory.stepInfo{end}.leakage;
          mol.currentData.acceptor = mol.currentData.acceptor + oldLeakage*mol.currentData.donor;
          mol.currentData.fret = mol.currentData.acceptor./(mol.currentData.acceptor + mol.currentData.donor);
          mol.leakage = 0;
        case 'firstAndLastRemoved'
          
        case 'singleFramesAnalyzed'
        case 'originalData'
          disp('shouldnothappen - moleculeData.m - undoLastStep');
          return;
        case 'redFramesRemoved'
        case 'zeroAdjusted'
                  oldAcceptorZero = mol.analysisStepHistory.stepInfo{end}.acceptorZeroAdjustment;
        oldDonorZero = mol.analysisStepHistory.stepInfo{end}.donorZeroAdjustment;
                  mol.currentData.donor = mol.currentData.donor + oldDonorZero;
        mol.currentData.acceptor = mol.currentData.acceptor + oldAcceptorZero;
        mol.currentData.fret = mol.currentData.acceptor./(mol.currentData.acceptor + mol.currentData.donor);
                mol.acceptorZero =  0;
      mol.donorZero = 0;
      end
      
      if isfield(dataUpdates, 'rows')
        dataFieldNames = fieldnames(dataUpdates.previousValues);
        for i = 1:size(dataFieldNames,1)
          for j = 1:length(dataUpdates.rows)
            mol.currentData.(dataFieldNames{i})(dataUpdates.rows(j)) = dataUpdates.previousValues.(dataFieldNames{i})(j);
          end
        end
      end
      mol.analysisStepHistory.changesToData(end) = [];
      mol.analysisStepHistory.stepInfo(end) = [];
      mol.analysisStepHistory.stepsTaken(end) = [];
      mol.currentAnalysisStep = mol.analysisStepHistory.stepsTaken{end};
      end
    end
    function saveIgorFileOfMolData(mol, saveFolder)
      nwaves = size(mol.analysisStepHistory.stepsTaken,2);
%       waveNamesAcceptor{1} = 'acceptor_final';
%       waveNamesDonor{1} = 'donor_final';
%       waveNamesFret{1} = 'FRET_final';
%       waveNamesClusteredSignal{1} = 'clustered_final';
%       dataToPrint = nan(length(mol.originalData.donor), (nwaves*3+1));
%       if isfield(mol.currentData, 'clusteredSignal')
%       dataToPrint(:,1) = mol.currentData.clusteredSignal;
%       end
%       dataToPrint(:,2) = mol.currentData.fret;
%       dataToPrint(:,3) = mol.currentData.acceptor;
%       dataToPrint(:,4) = mol.currentData.donor;
%       
%       dataPrintIdx = 5;
%       waveDeclarations = ['WAVES/D ',waveNamesClusteredSignal{1}, ', ', waveNamesFret{1}, ', ', waveNamesAcceptor{1}, ', ', waveNamesDonor{1}, ','];
      dataNames = { 'fret','clusteredSignal', 'acceptor', 'donor'};
      %%% new added 9/28
      for i = 1:4
        dataToPrint.(dataNames{i}) = nan(length(mol.originalData.donor), nwaves);
        waveNamesAcceptor = cell.empty;
        waveNamesDonor = cell.empty;
        waveNamesFret = cell.empty;
        waveNamesClusteredSignal = cell.empty;
        dataIdx.(dataNames{i}).TF = zeros(nwaves, 1, 'logical');
        dataIdx.(dataNames{i}).ix = 1;
      end
      %%% end of new added 9/28
%       formatSpecForData = ['%f %f %i %i '];
%       for i = 1:nwaves-1
%         backwardsIdx = nwaves - i;
%         stepName = mol.analysisStepHistory.stepsTaken{backwardsIdx};
%         if ~isempty(mol.analysisStepHistory.changesToData{backwardsIdx})
%           nfields = length(fieldnames(mol.analysisStepHistory.changesToData{backwardsIdx}.previousValues));
%           if isfield(mol.analysisStepHistory.changesToData{backwardsIdx}.previousValues, 'clusteredSignal')
%             nfields = 4;
%           else
%             nfields = 3;
%           end
%           switch stepName
%             case 'postBleachRemoval'
%               stepName = 'filtered';
%           end
%           if ismember(['acceptor_',stepName], waveNamesAcceptor)
%             [~, c] = find(contains(waveNamesAcceptor, ['acceptor_',stepName]));
%             waveNamesAcceptor{end+1} = ['acceptor_', stepName, num2str(size(c,2))];
%             waveNamesDonor{end+1} = ['donor_', stepName, num2str(size(c,2))];
%             waveNamesFret{end+1} = ['FRET_', stepName, num2str(size(c,2))];
%             if nfields == 4
%               waveNamesClusteredSignal{end+1} = ['clustered_', stepName, num2str(size(c,2))];
%               waveDeclarations = [waveDeclarations, ' ', waveNamesClusteredSignal{end}, ', ', waveNamesFret{end}, ', ',waveNamesAcceptor{end}, ', ', waveNamesDonor{end}, ','];
%             else
%               waveDeclarations = [waveDeclarations, ' ', waveNamesFret{end}, ', ',waveNamesAcceptor{end}, ', ', waveNamesDonor{end}, ','];
%             end
%           else
%             waveNamesAcceptor{end+1} = ['acceptor_', stepName];
%             waveNamesDonor{end+1} = ['donor_', stepName];
%             waveNamesFret{end+1} = ['FRET_', stepName];
%             if nfields == 4
%               waveNamesClusteredSignal{end+1} = ['clustered_', stepName];
%               waveDeclarations = [waveDeclarations, ' ', waveNamesClusteredSignal{end}, ', ', waveNamesFret{end}, ', ',waveNamesAcceptor{end}, ', ', waveNamesDonor{end}, ','];
%             else
%               waveDeclarations = [waveDeclarations, ' ', waveNamesFret{end}, ', ',waveNamesAcceptor{end}, ', ', waveNamesDonor{end}, ','];
%             end
%           end
%           if nfields == 4
%             p = 1;
%             formatSpecForData = [formatSpecForData, '%f %f %i %i '];
%           else
%             p = 2;
%             formatSpecForData = [formatSpecForData, '%f %i %i '];
%           end
%           for q = p:4
%             dataToEdit = dataToPrint(:,dataPrintIdx-nfields);
%             dataToEdit(mol.analysisStepHistory.changesToData{backwardsIdx}.rows) = ...
%               mol.analysisStepHistory.changesToData{backwardsIdx}.previousValues.(dataNames{q});
%             dataPrintIdx = dataPrintIdx + 1;
%             dataToPrint(:,dataPrintIdx) = dataToEdit;
%           end
%         end
%       end
      waveIndex = zeros(nwaves, 1, 'logical');
      for n = 1:nwaves
        stepName = mol.analysisStepHistory.stepsTaken{n};
        switch stepName
          case 'postBleachRemoval'
              stepName = 'filtered';
          case 'originalData'
            stepName = 'raw';
        end
        %adding an index to end of step name if there are multiple steps
        %with the same name
        TF = matches(['acceptor_',stepName], waveNamesAcceptor);
        if TF == true
          idx = 1;
        else 
          idx = [];
        end
        while TF == true
          idx = idx + 1;
          newStepName = [stepName, num2str(idx)];
          TF = matches(['acceptor_',newStepName], waveNamesAcceptor);
        end
        %adding wave name to lists
        waveNamesAcceptor{end+1} = ['acceptor_', stepName, num2str(idx)];
        waveNamesDonor{end+1} = ['donor_', stepName, num2str(idx)];
        waveNamesFret{end+1} = ['FRET_', stepName, num2str(idx)];
        waveNamesClusteredSignal{end+1} = ['clustered_', stepName, num2str(idx)];
        if ~isempty(mol.analysisStepHistory.changesToData{n})
          waveIndex(n) = true;
        end
      end
      
      
      
      for i = 1:4
        if isfield(mol.currentData, dataNames{i})
        dataToPrint.(dataNames{i})(:,nwaves) = mol.currentData.(dataNames{i});
        dataIdx.(dataNames{i}).TF(nwaves) = true;
        dataIdx.(dataNames{i}).ix = nwaves;
        end
%         if isfield(mol.originalData, dataNames{i})
%         dataToPrint.(dataNames{i})(:,1) = mol.originalData.(dataNames{i});
%         end
      end
      
      n = nwaves; 
      while n > 1
        if waveIndex(n) == true
          for i = 1:4
            if isfield(mol.analysisStepHistory.changesToData{n}.previousValues, dataNames{i})
              indx = dataIdx.(dataNames{i}).ix;
            dataToEdit = dataToPrint.(dataNames{i})(:,indx);
            dataToEdit(mol.analysisStepHistory.changesToData{n}.rows) = ...
              mol.analysisStepHistory.changesToData{n}.previousValues.(dataNames{i});
            dataToPrint.(dataNames{i})(:,n-1) = dataToEdit;
            dataIdx.(dataNames{i}).TF(n-1) = true;
            dataIdx.(dataNames{i}).ix = n - 1;
%             elseif matches(mol.analysisStepHistory.stepsTaken{n}, 'postClustering') && matches(dataNames{i}, 'clusteredSignal')
%               indx = dataIdx.(dataNames{i}).ix;
%               dataToEdit = dataToPrint.(dataNames{i})(:,indx);
%               dataToEdit(mol.analysisStepHistory.changesToData{n}.rows) = ...
%                 mol.analysisStepHistory.changesToData{n}.previousValues.(dataNames{i});
%               dataToPrint.(dataNames{i})(:,n-1) = dataToEdit;
%               dataIdx.(dataNames{i}).TF(n-1) = true;
%               dataIdx.(dataNames{i}).ix = n - 1;
            elseif matches(mol.analysisStepHistory.stepsTaken{n}, {'zeroAdjustedForDonor', 'zeroAdjustedForAcceptor'})
              if matches(mol.analysisStepHistory.stepInfo{n}.dataName, dataNames{i})
                indx = dataIdx.(dataNames{i}).ix;
                dataToEdit = dataToPrint.(dataNames{i})(:,indx);
                dataToEdit = dataToEdit - mol.analysisStepHistory.stepInfo{n}.value;
                dataToPrint.(dataNames{i})(:,n-1) = dataToEdit;
                dataIdx.(dataNames{i}).TF(n-1) = true;
                dataIdx.(dataNames{i}).ix = n - 1;
              end
            end
            
          end
        end
        n = n-1;
      end
      waveDeclarations = ['WAVES/D ']; formatSpecForData = char.empty;
      %waveDeclarations = ['WAVES/D ',waveNamesClusteredSignal{1}, ', ', waveNamesFret{1}, ', ', waveNamesAcceptor{1}, ', ', waveNamesDonor{1}, ','];
      dataToPrint.allData = nan(length(mol.originalData.donor), (nwaves*4+1));
      bigIdx = 1; colorsForPlot = {'m', 'b', 'r', 'g'};
                f1 = figure('visible', 'off'); f2 = figure('visible', 'off');
                [a,~] = find(dataIdx.acceptor.TF == true);
          tL = tiledlayout(f1, length(a),1); 
          [a,~] = find(dataIdx.fret.TF == true);
          tL2 = tiledlayout(f2, length(a),1);
          title(tL,'Donor and Acceptor Analysis Steps','interpreter','none');
          title(tL2,'Fret and Clustered Signal Analysis Steps','interpreter','none');
                          plot1Started = false;
                plot2Started = false;
      for i = 1:4
        [a,~] = find(dataIdx.(dataNames{i}).TF == true);
        if ~isempty(a)
          switch dataNames{i}
            case 'acceptor'
              waveNames = waveNamesAcceptor(a);
              plottingStyle = 1;
            case 'donor'
              waveNames = waveNamesDonor(a);
              plottingStyle = 1;
            case 'fret'
              waveNames = waveNamesFret(a);
              plottingStyle = 2;
            case 'clusteredSignal'
              waveNames = waveNamesClusteredSignal(a);
              plottingStyle = 2;
          end
          

          for n = 1:length(a)
            waveDeclarations = [waveDeclarations, waveNames{n}, ', '];
            dataToPrint.allData(:,bigIdx) = dataToPrint.(dataNames{i})(:,a(n));
            bigIdx = bigIdx + 1;
            formatSpecForData = [formatSpecForData, '%f '];
            switch plottingStyle
              case 1
                figure(f1);
                axName = ['plotStyle', num2str(plottingStyle), num2str(a(n))];
                if plot1Started == false
                  ax.(axName) = nexttile(tL);
                  
                  if n == 1
                    xlim = [1, length(dataToPrint.(dataNames{i})(:,a(n)))];
                  end
                  if n == length(a)
                    plot1Started = true;
                  end
                end
                t = 1:length(dataToPrint.(dataNames{i})(:,a(n)));
                t = t';
                plot(ax.(axName), t, dataToPrint.(dataNames{i})(:,a(n)), colorsForPlot{i});
                title(ax.(axName), mol.analysisStepHistory.stepsTaken{a(n)},'interpreter','none');
                  ax.(axName).XLim = xlim;
                hold(ax.(axName), 'on');
                
                
              case 2
                figure(f2);
                axName = ['plotStyle', num2str(plottingStyle), num2str(a(n))];
                if plot2Started == false
                  ax.(axName) = nexttile(tL2);
                  
                  if n == 1
                    xlim = [1, length(dataToPrint.(dataNames{i})(:,a(n)))];
                  end
                  if n == length(a)
                    plot2Started = true;
                  end
                end
                t = 1:length(dataToPrint.(dataNames{i})(:,a(n)));
                t = t';
                plot(ax.(axName), t, dataToPrint.(dataNames{i})(:,a(n)), colorsForPlot{i});
                title(ax.(axName), mol.analysisStepHistory.stepsTaken{a(n)},'interpreter','none');
                  ax.(axName).XLim = xlim;
                hold(ax.(axName), 'on');
                
            end

          end
        end

        
      end
      
      exportgraphics(f1, [saveFolder, '\images\DonorAcceptor_stepsPlot.png']);
      exportgraphics(f2, [saveFolder, '\images\FRET_stepsPlot.png']);
      


      
      
      
                waveDeclarations = waveDeclarations(1:(end-1));
      waveDeclarations = string(waveDeclarations);
      firstThreeLines = ["IGOR", waveDeclarations,"BEGIN"];
      
      
      formatSpecForData = [formatSpecForData, '\n'];
      dataToPrint.allData(:,bigIdx:end) = [];
      d = dataToPrint.allData;
      d = d';

      
      % now save the waves
      saveName = [saveFolder, '\trace', num2str(mol.traceNum), 'mol', num2str(mol.moleculeNumber), 'Data.itx'];
      fid = fopen(saveName,'w');
      fprintf(fid, '%s\n',firstThreeLines);
      fprintf(fid, formatSpecForData, d);
      fprintf(fid,'END\n');


        end

      end

 
  
  methods (Static) %functions used by this class that are saved in separate files
    %all of these functions must be saved in the same folder as this
    %script, and the folder must start with '@' and be followed by the name
    %of the class. In this case, the folder is called @moleculeData
    
    %note to self: later on, move all static functions here and create
    %separate files for them. 
    TF = checkForAllNanValues(data);
    sections = separateOutSections(clusteredSignal);
    varargout = locateSingleFrameEvents(sections);
  end


  
end