classdef bleachRemovalResults 
  properties
    method
    threshold
    windowSize
    smoothFunction
    applyMaxThreshTF
    maxThresh
    dataToUse
    padding
    numConsecutiveFrames
    actionToTake
    maxLengthOfBlink
    conditions
    replacementValues = [];
    numberOfRuns = 1;
    greenEndFrame
    greenStartFrame
    redFramesSmoothed
    redFrameData
  end
  properties (Transient)
    smoothedSum
    bleachIndex
    results % struct with fields: donor, acceptor, fret
    
  end
  
  %%%% different data filtering options:
  %%% REMOVE:
  %%% 1. if dataUsed < threshold
  %%% 2. if dataUsed < threshold or dataUsed > maxThresh
  %%% 
  %%% REMOVE IF:
  %%% 3. if dataUsed < threshold and 'Sections Before And After Are Bleached';
  %%% 4. if dataUsed < threshold and 'Section is at begining or end'
  %%% 5. if dataUsed < threshold and 'Red Frames Are Bleached' and section
  %%% is next to red frame section
  %%% 6. if dataUsed < threshold and length(section) < minLength
  %%% 7. if dataUsed < threshold and length(section) > minLength
  %%% 
  %%%'Specify min length of sections to be removed'; 'Section is at begining or end'; ...
               % 'Red Frames Are Bleached'
               
               
  %%% Rewrite constructor function like this:
  % switch filterType, case "Threshold", case 'Whole Molecule'
  %case 'Whole Molecule'
  % check TF for conditions
  
  methods
    function bleachData = bleachRemovalResults(originalData, bleachSettings, method, options)
      arguments
        originalData
        bleachSettings
        method = 'threshold';
        options.redFrameData
      end


      % gather filtering information from main app
      dataNames = fieldnames(bleachSettings);
      for i = 1:size(dataNames,1)
        bleachData.(dataNames{i}) = bleachSettings.(dataNames{i});
      end
      if isfield(bleachSettings, 'method')
        bleachData.method = bleachSettings.method;
      else
        bleachData.method = method;
      end
      
      
      % define which data to use 
      switch bleachData.dataToUse
        case 'Sum Acceptor Donor'
          data = originalData.donor + originalData.acceptor;
        case {'Acceptor', 'Donor', 'Fret'}
          dataName = lower(bleachData.dataToUse);
          data = originalData.(dataName);
        case {'Donor AND Acceptor', 'Donor OR Acceptor'}
          bleachData.numberOfRuns = 2;
        case 'all' %this is when an option was selected to remove the whole trace
          data = nan(length(originalData.donor));
      end
      
      
      % get index and replacement values if they apply
      switch bleachData.numberOfRuns
        case 1
          %smooth the data
          bleachData.smoothedSum = smoothData(bleachData, data, isnan(originalData.fret));
          if isfield(options, 'redFrameData') %smooth red frame data too if needed
            options.redFrameData((bleachData.greenStartFrame-bleachData.maxLengthOfBlink):(bleachData.greenEndFrame+bleachData.maxLengthOfBlink)) = nan;
            bleachData.redFramesSmoothed = smoothData(bleachData, options.redFrameData, isnan(options.redFrameData));
          end
          [bleachData.bleachIndex, bleachData.replacementValues] = applySettingsForBleachIndex(bleachData, originalData, ...
            bleachData.smoothedSum, bleachData.threshold, bleachData.replacementValues);
        case 2
          nanIdx = isnan(originalData.fret);
          dataNames = {'acceptor', 'donor'};
          thresholds = [bleachData.threshold, bleachData.maxThresh];
          twoIndexes = zeros(length(originalData.fret),2,'logical');
          bleachData.smoothedSum = zeros(length(originalData.fret),2);

          for i = 1:2
            if isfield(options, 'redFrameData')
              options.redFrameData((bleachData.greenStartFrame-bleachData.maxLengthOfBlink):(bleachData.greenEndFrame+bleachData.maxLengthOfBlink)) = nan;
              nan2Idx = isnan(options.redFrameData);
              bleachData.redFramesSmoothed = smoothData(bleachData, options.redFrameData.(dataNames{i}), nan2Idx);
            end
            data = originalData.(dataNames{i});
            bleachData.smoothedSum(:,i) = smoothData(bleachData, data, nanIdx);
            [twoIndexes(:,i), bleachData.replacementValues] = applySettingsForBleachIndex(bleachData, originalData,...
              bleachData.smoothedSum(:,i), thresholds(i), bleachData.replacementValues);
          end
          switch bleachData.dataToUse
            case 'Donor AND Acceptor'
              bleachData.bleachIndex = and(twoIndexes(:,1), twoIndexes(:,2));
            case 'Donor OR Acceptor'
              bleachData.bleachIndex = or(twoIndexes(:,1), twoIndexes(:,2));
          end
          
      end
      

      bleachData.results = bleachData.applyBleachIndex(originalData, bleachData.bleachIndex, bleachData.replacementValues);
    end
    function dataSmoothed = smoothData(bleachData, data, nanIdx)
      % smooth the data
      switch bleachData.smoothFunction
        case 'average'
          smoothFcn = @(data, windowSize) movmean(data, windowSize, 'omitnan');
        case 'median'
          smoothFcn = @(data, windowSize) movmedian(data, windowSize, 'omitnan');
      end
      if bleachData.windowSize ~= 0
        dataSmoothed = smoothFcn(data, bleachData.windowSize);
      else
        dataSmoothed = data;
      end
      dataSmoothed(nanIdx) = nan;
    end
    
    function [Index, valuesForReplacement] = applySettingsForBleachIndex(bleachData, originalData, smoothSum, thresh, valuesForReplacement)
      % create an index of points to be removed or replaced based on
      % settings and method selected
      switch bleachData.method
        case 'removedByHand'
          Index = ones(length(originalData.donor), 1, 'logical');
        otherwise
          Index = bleachData.getBleachIndex(thresh, smoothSum, ...
            bleachData.applyMaxThreshTF, bleachData.maxThresh, bleachData.numConsecutiveFrames);
          Index = bleachData.padEndsOfBleachedSections(Index, bleachData.padding);
          if ~isempty(bleachData.conditions)
            switch bleachData.conditions
              case {'Sections Before And After Are Bleached'}
                Index = analyzeAdjacentSections(bleachData, Index, originalData);
              case {'Specify min length of sections to be removed', 'Specify max length of sections to be replaced'}
                valuesForReplacement = applyConditionals(bleachData, originalData, Index, valuesForReplacement);
              case {'Section is at begining or end'}
                Index = selectSectionsAtBeginingOrEnd(bleachData, Index, originalData);
              case {'Red Frames Are Bleached'}
                dataForSelectSections = originalData;
                dataNames = {'fret'; 'donor'; 'acceptor'};
                for i = 1:size(dataNames,1)
                  dataForSelectSections.(dataNames{i})(1:bleachData.greenStartFrame) = nan;
                  if bleachData.greenEndFrame ~= inf
                    dataForSelectSections.(dataNames{i})(bleachData.greenEndFrame:end) = nan;
                  end
                end
                Index = selectSectionsAtBeginingOrEnd(bleachData, Index, dataForSelectSections);
                [idxRowStart, idxRowEnd] = bleachData.separateIndexIntoSections(Index);
                redFramesInBegining = bleachData.redFramesSmoothed(1:(bleachData.greenStartFrame-bleachData.maxLengthOfBlink));
                redFramesInEnd = bleachData.redFramesSmoothed((bleachData.greenEndFrame+bleachData.maxLengthOfBlink):end);
                redFramesInBegining(isnan(redFramesInBegining)) = thresh - 10;
                redFramesInEnd(isnan(redFramesInEnd)) = thresh - 10;
                if bleachData.greenEndFrame ~= inf
                  if ~all(redFramesInEnd < thresh) && length(idxRowStart) > 2
                    Index(idxRowStart(end-1):idxRowEnd(end-1)) = false;
                  end
                end
                if ~all(redFramesInBegining < thresh)&& length(idxRowStart) > 3
                  Index(idxRowStart(2):idxRowEnd(2)) = false;
                end

                
            end
          end
      end
    end
    function bleachIndex = padEndsOfBleachedSections(bleachData, bleachIndex, paddingAmount)
      totalLength = length(bleachIndex);
      [idxRowStart, idxRowEnd] = bleachData.separateIndexIntoSections(bleachIndex, 'bleachedSectionsOnly');
      for i = 1:length(idxRowStart)
        frontRowsToPad = (idxRowStart(i) - paddingAmount):idxRowStart(i);
        frontRowsToPad = frontRowsToPad(frontRowsToPad >= 1);
        bleachIndex(frontRowsToPad) = true;
        backRowsToPad = idxRowEnd(i):(idxRowEnd(i) + paddingAmount);
        backRowsToPad = backRowsToPad(backRowsToPad <= totalLength);
        bleachIndex(backRowsToPad) = true;
      end
    end
    function realNewBleachIndex = selectSectionsAtBeginingOrEnd(bleachData, bleachIndex, originalData)
            newBleachIndex = double(bleachIndex);
      nanIndex = isnan(originalData.acceptor);
      newBleachIndex(nanIndex) = 3;
      realNewBleachIndex = zeros(size(newBleachIndex), 'logical');
      [idxRowStart, idxRowEnd] = bleachData.separateIndexIntoSections(newBleachIndex);
      if newBleachIndex(idxRowStart(1)) == 3 %found red frames
        if length(idxRowStart) > 1 && newBleachIndex(idxRowStart(2)) == true
          realNewBleachIndex(idxRowStart(2):idxRowEnd(2)) = true;
        end
      elseif newBleachIndex(idxRowStart(1)) == 1  %if first section is bleached
        realNewBleachIndex(idxRowStart(1):idxRowEnd(1)) = true;
      end
      lastIndexNum = length(idxRowEnd);
      if newBleachIndex(idxRowEnd(lastIndexNum)) == 3 %found red frames
        if length(idxRowEnd) > 1 && newBleachIndex(idxRowStart(lastIndexNum-1)) == true
          realNewBleachIndex(idxRowStart(lastIndexNum-1):idxRowEnd(lastIndexNum-1)) = true;
        end
      elseif newBleachIndex(idxRowEnd(lastIndexNum)) == 1  %if first section is bleached
        realNewBleachIndex(idxRowStart(lastIndexNum):idxRowEnd(lastIndexNum)) = true;
      end
    end
    function newBleachIndex = analyzeAdjacentSections(bleachData, bleachIndex, originalData)
      %
      newBleachIndex = double(bleachIndex);
      nanIndex = isnan(originalData.acceptor);
      newBleachIndex(nanIndex) = 3;
      [idxRowStart, idxRowEnd] = bleachData.separateIndexIntoSections(newBleachIndex);
      for i = 1:size(idxRowStart,1)
        %first test to see if this section is a section that was already
        %removed in a previous step. If this is the case, then the values
        %for the data in this section will be nan
        testForNanValue = originalData.acceptor(idxRowStart(i));
        testForBleachSection = bleachIndex(idxRowStart(i));
        if testForBleachSection == true %|| isnan(testForNanValue) 
          continue;
        end
        %now, if this section is not nan, then test the sections before and
        %after to see if they are bleached sections. Goal is to only keep
        %sections where the adjacent sections are bleached sections
        %test the section before and after, if they are both bleach
        %sections, then this section stays, otherwise, this section is
        %also a bleach section
        %additionally, if this is the first or last section, then it is
        %automatically turned into a bleach section
%         if i ~= 1 && i ~= size(idxRowStart,1)
          if idxRowStart(i) ~= 1 && idxRowEnd(i) ~= length(bleachIndex)
            testValueSectionBefore = newBleachIndex(idxRowStart(i-1));
            testValueSectionAfter = newBleachIndex(idxRowStart(i+1));
            %if both sections before and after are in the bleach Index,
            %then keep this section, otherwise, add it to the bleach
            %sections
            if testValueSectionBefore == 1 && testValueSectionAfter == 1
%               %also test to see if the sections before are in the red laser
%               %frames
%               testNanValueSectionBefore = isnan(originalData.acceptor(idxRowStart(i-1)));
%               testNanValueSectionAfter = isnan(originalData.acceptor(idxRowStart(i+1)));
%               if testNanValueSectionBefore == false && testNanValueSectionAfter == false
                newBleachIndex(idxRowStart(i):idxRowEnd(i)) = 0;
              else
                newBleachIndex(idxRowStart(i):idxRowEnd(i)) = 1;
              end
%             end
          end
%         else
%           newBleachIndex(idxRowStart(i):idxRowEnd(i)) = 0;
%         end
      end
      newBleachIndex(nanIndex) = 0;
      newBleachIndex = logical(newBleachIndex);
    end
    
    
    function [idxRowStart, idxRowEnd] = separateIndexIntoSections(bleachData, index, whichSections)
      arguments
        bleachData %#ok
        index
        whichSections = 'all';
      end
      index = double(index);
      b = [index(2:end); nan];
      [idxRowEnd, ~] = find(b ~= index);
      c = [nan; index(1:end-1)];
      [idxRowStart, ~] = find(c~= index);
      switch whichSections
        case 'bleachedSectionsOnly'
          index = logical(index);
          idxRowStart = idxRowStart(index(idxRowStart) == true);
          idxRowEnd = idxRowEnd(index(idxRowEnd) == true);
        case 'nonBleachSections'
          index = logical(index);
          idxRowStart = idxRowStart(index(idxRowStart) == false);
          idxRowEnd = idxRowEnd(index(idxRowEnd) == false);
      end
    end
    
    function replacementValues = applyConditionals(bleachData, originalData, index, replacementValues)
      arguments
        bleachData
        originalData
        index
        replacementValues = struct.empty;
      end
      dataFields = {'donor', 'acceptor', 'fret'};
      [idxRowStart, idxRowEnd] = bleachData.separateIndexIntoSections(index, 'bleachedSectionsOnly');
      for i = 1:3
        replacementValues.(dataFields{i}) = originalData.(dataFields{i});
      end
      actionsThatHaveConditionalReplacement = {'Replace IF', 'Remove IF'};
      actionsThatNeedReplacementValuesCalculated = {'Replace', 'Replace IF'};
      actionsThatHaveSpecificReplacementValues = {'Replace WITH', 'Replace WITH value IF'};
      requirement = bleachData.maxLengthOfBlink;
      switch bleachData.actionToTake
        case {'Replace IF', 'Remove IF'}
          requirementForReplacement = requirement + bleachData.padding*2;
        case 'Replace WITH'
          value = requirement;
      end
      for i = 1:length(idxRowStart)
        sectionRows = idxRowStart(i):idxRowEnd(i);
        % addressing actionsThatHaveConditionalReplacement
        if ismember(bleachData.actionToTake, actionsThatHaveConditionalReplacement)
          if length(sectionRows) >= requirementForReplacement
            for j = 1:3
              replacementValues.(dataFields{j})(sectionRows) = nan;
            end
            continue
          end
        end % end of addressing actionsThatHaveConditionalReplacement
        
        % addressing actionsThatNeedReplacementValuesCalculated
        if ismember(bleachData.actionToTake, actionsThatNeedReplacementValuesCalculated)
          rowBeforeBleachSection = idxRowStart(i) - 1;
          rowAfterBleachSection = idxRowEnd(i) +1;
          intensityData = originalData.donor;
          TF = [isnan(intensityData(rowBeforeBleachSection)), ...
            isnan(intensityData(rowAfterBleachSection))];
          if all(TF)
            for j = 1:3
              replacementValues.(dataFields{j})(sectionRows) = nan;
              continue
            end
          elseif any(TF)
            b = [rowBeforeBleachSection, rowAfterBleachSection];
            rowToUseToReplaceDataValues = b(~TF);
            for j = 1:2
              replacementValues.(dataFields{j})(sectionRows) = originalData.(dataFields{j})(rowToUseToReplaceDataValues);
            end
          else
            for j = 1:2
              dataBeforeSection = originalData.(dataFields{j})(rowBeforeBleachSection);
              dataAfterSection = originalData.(dataFields{j})(rowAfterBleachSection);
              avgValue = (dataBeforeSection + dataAfterSection)/2;
              replacementValues.(dataFields{j})(sectionRows) = avgValue;
            end
          end
          replacementValues.fret(sectionRows) = replacementValues.acceptor(sectionRows)./(replacementValues.acceptor(sectionRows) + ...
            replacementValues.donor(sectionRows));
        end % end of addressing actionsThatNeedReplacementValuesCalculated
        % addressing actionsThatHaveSpecificReplacementValues
        if ismember(bleachData.actionToTake, actionsThatHaveSpecificReplacementValues)
          dataName = lower(bleachData.dataToUse);
          TF = matches(dataFields, dataName);
          for j = 1:3
            switch TF(j)
              case true %replace with value
                replacementValues.(dataFields{j})(sectionRows) = value;
              case false %replacement values same as original values
                replacementValues.(dataFields{j})(sectionRows) = originalData.(dataFields{j})(sectionRows);
            end
          end
          if TF(3) == false
            replacementValues.fret(sectionRows) = replacementValues.acceptor(sectionRows)./(replacementValues.acceptor(sectionRows) + ...
              replacementValues.donor(sectionRows));
          end
        end % end of addressing actionsThatHaveSpecificReplacementValues
      end %end of loop through sections
    end

  end
  methods (Static)
    function resultsAfterBleachRemoval = applyBleachIndex(originalData, index, replacementValues)
      dataFields = {'donor', 'acceptor', 'fret'};
      if isempty(replacementValues)
        for i = 1:3
          result = originalData.(dataFields{i});
          result(index)= nan;
          resultsAfterBleachRemoval.(dataFields{i}) = result;
        end
      else
        for i = 1:3
          result = originalData.(dataFields{i});
          newValues = replacementValues.(dataFields{i});
          result(index)= newValues(index);
          resultsAfterBleachRemoval.(dataFields{i}) = result;
        end
      end
    end

    function bleachIndex = getBleachIndex(threshold, dataToUse, ...
        applyMaxThreshTF, maxThresh, numConsecutiveFramesRequiredForRemoval)
      
      
      
      
      bleachIndex = dataToUse < threshold;
      if applyMaxThreshTF == true
        bleachIndexMaxThresh = dataToUse > maxThresh;
        bleachIndex = or(bleachIndex, bleachIndexMaxThresh);
      end
      %rewrite section below using separateIndexIntoSections function
      if numConsecutiveFramesRequiredForRemoval > 0
      firstBleachSection = find(bleachIndex == true, 1);
      if firstBleachSection == 1
        numBleachFrames = find(bleachIndex == false, 1) - 1;
        if numBleachFrames <= numConsecutiveFramesRequiredForRemoval
          bleachIndex(1:numBleachFrames) = false;
        end
      end
      i=2;
      while i < length(bleachIndex)
        if bleachIndex(i) == true
          if bleachIndex(i+1) == false
            temp = bleachIndex((i+1):end);
            numFramesUntilNextBleachSection = find(temp == true,1) - 1;
            if numFramesUntilNextBleachSection <= numConsecutiveFramesRequiredForRemoval
              bleachIndex(i:(i+numFramesUntilNextBleachSection)) = false;
            end
          end
          if bleachIndex(i-1) == false
            temp = bleachIndex(i:end);
            numBleachFrames = find(temp == false,1)-1;
            if numBleachFrames <= numConsecutiveFramesRequiredForRemoval
              bleachIndex(i:(i+numBleachFrames)) = false;
            end
          end
        end
        i = i+1;
      end
      end
    end


  end
end